<?php

namespace EventPortal\Domain\Form;

use Closure;
use EventPortal\Domain\Agenda\SessionErrors;
use EventPortal\Domain\Component\Input;
use EventPortal\Domain\Component\Section;
use EventPortal\Domain\Component\Step;
use EventPortal\Domain\Util\RelationshipType;
use EventPortal\Domain\Util\ValidationRules;
use EventPortal\Exceptions\NoSaveHandlers;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Pipeline\Pipeline;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\ValidationException;

class Form
{
    /** @var string */
    public $name;

    /** @var array */
    public $blocks = [];

    /** @var array */
    public $actions = [];

    /** @var null|callable */
    public $primaryAction;

    /** @var Model */
    private $model;

    /** @var array */
    private $context = [];

    /** @var array */
    private $saveHandlers;

    /** @var bool */
    private $useAlternativeName = false;

    /** @var Validator */
    private $validator;

    /** @var bool */
    private $database = false;

    public function __construct(string $name)
    {
        $this->name = $name;
    }

    /**
     * @param callable|string $fields
     */
    public function addFields($fields): self
    {
        $this->blocks[] = $fields;

        return $this;
    }

    public function addActions(callable $actions): self
    {
        $this->actions = $actions;

        return $this;
    }

    public function addPrimaryAction(callable $primaryAction): self
    {
        $this->primaryAction = $primaryAction;

        return $this;
    }

    /**
     * @param mixed ...$models
     */
    public function context(...$models): self
    {
        $keyedModels = [];

        foreach ($models as $model) {
            $keyedModels[$model::KEY] = $model;
        }

        $this->context = $keyedModels;

        if ($this->model) {
            $this->context = array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $this;
    }

    public function model(Model $model): self
    {
        $this->model = $model;

        if ($this->context) {
            $this->context = array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $this;
    }

    public function validate(array $input): self
    {
        $this->resolveWithoutHydration();

        $this->validator = Validator::make($input, ValidationRules::for($this->blocks)->get());

        if ($this->useAlternativeName) {
            return $this->validateUsingAlternativeNames();
        }

        $this->validator->validate();

        return $this;
    }

    public function render(): array
    {
        $this->resolve();

        $data = [
            'actions' => $this->actions,
            'content' => $this->blocks,
        ];

        if ($this->primaryAction !== null) {
            $data['onEnter'] = $this->primaryAction;
        }

        return $data;
    }

    public function addHandler(callable $saveHandler): self
    {
        $this->saveHandlers[] = $saveHandler;

        return $this;
    }

    public function addHandlers(array $saveHandlers): self
    {
        foreach ($saveHandlers as $saveHandler) {
            $this->saveHandlers[] = $saveHandler;
        }

        return $this;
    }

    public function useAlternativeNames(): self
    {
        $this->useAlternativeName = true;

        return $this;
    }

    /**
     * @return mixed
     */
    public function handle(array $input)
    {
        throw_if($this->saveHandlers === [] || $this->saveHandlers === null, NoSaveHandlers::class);

        $payload = Payload::make($input)
            ->withModel($this->model)
            ->withContext($this->context);

        $this->reset();

        return app(Pipeline::class)
            ->send($payload)
            ->through($this->saveHandlers)
            ->thenReturn();
    }

    private function reset(): void
    {
        $this->blocks = [];
        $this->actions = [];
        $this->primaryAction = null;
    }

    private function resolveWithoutHydration(): void
    {
        $this->resolve(false);
    }

    private function resolve(bool $hydrate = true): void
    {
        $this->resolveActions();
        $this->resolveSubmit();
        $this->resolveBlocks($hydrate);
    }

    private function resolveActions(): void
    {
        if ($this->actions instanceof Closure === false) {
            return;
        }

        $this->actions = app()->call($this->actions, $this->getContext());
    }

    private function resolveSubmit(): void
    {
        if ($this->primaryAction instanceof Closure === false) {
            return;
        }

        $this->primaryAction = app()->call($this->primaryAction, $this->getContext());
    }

    /**
     * Some documentation.
     */
    private function resolveBlocks(bool $hydrate): void
    {
        $this->blocks = collect($this->blocks)
            ->flatten(1)
            ->map(function ($block) {
                if ($block instanceof Closure) {
                    return app()->call($block, $this->getContext());
                }

                return $block;
            })
            ->map(function ($item) {
                if ($item instanceof Step) {
                    return $item;
                }

                if ($item instanceof Section) {
                    $item->children = $item->flatChildren();

                    return Step::make()->fields([$item]);
                }

                $containsSections = collect($item)
                    ->filter(function ($item) {
                        return $item instanceof Section;
                    })
                    ->isNotEmpty();

                $sections = Section::make()->fields([$item]);

                if ($containsSections) {
                    $sections = collect($item)->flatten()->toArray();

                    return Step::make()->fields($sections);
                }

                return Step::make()->fields([$sections]);
            })
            ->when($hydrate, function (Collection $steps) {
                return $this->hydrate($steps);
            })
            ->toArray();
    }

    private function hydrate(Collection $steps): Collection
    {
        return $steps->each(function (Step $step) {
            return collect($step->children)->each(function (Section $section): void {
                collect($section->flatChildren())->each(function (Input $input): void {
                    if ($this->useAlternativeName) {
                        $input->useAlternativeName();
                    }

                    $input->value($this->getValue($input));
                });
            });
        });
    }

    /**
     * @return mixed|null
     */
    private function getValue(Input $input)
    {
        if ($input->value !== null) {
            return $input->value;
        }

        $name = $input->name;

        if ($input->isStatus()) {
            $name = 'readable_status';
        }

        if ($input->isNotScoped()) {
            $value = optional($this->model)->$name;

            if ($input->hasOptions()) {
                return $input->getOptions($value);
            }

            return $value;
        }

        if ($this->model && RelationshipType::make($this->model, $input->scope)->isBelongsToMany()) {
            $data = $this->model->{$input->scope}->first();

            return $data->pivot->$name;
        }

        if ($this->isFromDatabase()) {
            return $this
                ->model
                ->{$input->scope}
                ->map(static function ($data) use ($input) {
                    if ($input->label !== $data->question->question) {
                        return null;
                    }

                    $input->value($data->value);

                    return $input->getValue();
                })
                ->filter()
                ->first();
        }

        return object_get($this->model, $input->scopeKey());
    }

    public function getValidator(): \Illuminate\Validation\Validator
    {
        return $this->validator;
    }

    private function validateUsingAlternativeNames(): self
    {
        try {
            $this->validator->validate();
        } catch (ValidationException $exception) {
            $errors = SessionErrors::swapDatabaseKeysForBrowserKeys($exception->errors());

            foreach ($errors as $key => $value) {
                $this->validator->errors()->add($key, $value);
            }

            throw new ValidationException($this->validator);
        }

        return $this;
    }

    private function getContext(): array
    {
        $parameters = $this->context;

        if ($this->model) {
            return array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $parameters;
    }

    public function fromDatabase(): self
    {
        $this->database = true;

        return $this;
    }

    public function isFromDatabase(): bool
    {
        return $this->database === true;
    }
}
