<?php

namespace EventPortal\Tests\Unit\Form;

use EventPortal\Domain\Delegate\Models\Delegate;
use EventPortal\Domain\Event\Models\Event;
use EventPortal\Domain\Form\Payload;
use EventPortal\Domain\Speaker\Models\Speaker;
use EventPortal\Tests\TestCase;
use Illuminate\Foundation\Testing\RefreshDatabase;

class PayloadTest extends TestCase
{
    use RefreshDatabase;

    /** @test **/
    public function can_set_input()
    {
        $payload = Payload::make(['name' => 'Hello World!']);

        $this->assertEquals(['name' => 'Hello World!'], $payload->getInput());
    }

    /** @test **/
    public function can_get_input_by_key()
    {
        $payload = Payload::make([
            'name' => 'Hello World!',
            'other' => 'Another thing',
        ]);

        $this->assertEquals('Another thing', $payload->getInput('other'));
    }

    /** @test **/
    public function can_get_array_value_by_key()
    {
        $payload = Payload::make([
            'role_id' => ['label' => 'Super Admin', 'value' => 1],
            'other' => 'Another thing',
        ]);

        $this->assertEquals(1, $payload->getArrayValue('role_id'));
    }

    /** @test **/
    public function can_add_context()
    {
        $event = factory(Event::class)->make();

        $payload = Payload::make(['name' => 'Test'])->addContext('event', $event);

        $this->assertSameModel($event, $payload->getContext('event'));
    }

    /** @test **/
    public function can_add_context_multiple_times()
    {
        $event = factory(Event::class)->make();
        $other = ['random' => 'array'];

        $payload = Payload::make(['name' => 'Test'])
            ->addContext('event', $event)
            ->addContext('other', $other);

        $this->assertSameModel($event, $payload->getContext('event'));
        $this->assertSame($other, $payload->getContext('other'));
    }

    /** @test **/
    public function can_add_model()
    {
        $event = factory(Event::class)->make();

        $payload = Payload::make(['name' => 'Test'])->withModel($event);

        $this->assertSameModel($event, $payload->getModel());
    }

    /** @test **/
    public function can_replace_model()
    {
        $eventA = factory(Event::class)->make();
        $eventB = factory(Event::class)->make();

        $payload = Payload::make(['name' => 'Test'])->withModel($eventA);

        $payload->replaceModel($eventB);

        $this->assertSameModel($eventB, $payload->getModel());
    }

    /** @test **/
    public function can_add_to_input()
    {
        $payload = Payload::make(['name' => 'Test'])->addToInput(['email' => 'test@example.com']);

        $this->assertEquals('test@example.com', $payload->getInput('email'));
    }

    /** @test **/
    public function can_find_instance_model_or_in_context()
    {
        $event = factory(Event::class)->make();
        $delegate = factory(Delegate::class)->make();

        $payload = Payload::make(['name' => 'Test'])
            ->withModel($event)
            ->withContext(['delegate' => $delegate]);

        $this->assertNull($payload->find(Speaker::class));
        $this->assertSameModel($delegate, $payload->find(Delegate::class));
        $this->assertSameModel($event, $payload->find(Event::class));
    }
}
